import subprocess

from .. import Plugin

class ShellPlugin(Plugin):
    def exec(self, cmd):
        output = None
        error = None

        try:
            output = subprocess.check_output(cmd, stderr=subprocess.STDOUT, shell=True)
        except subprocess.CalledProcessError as e:
            error = e.output

        return [output, error]

# vim:sw=4:ts=4:et:




import logging
import json

from kafka import KafkaConsumer, KafkaProducer

from .. import Backend

class KafkaBackend(Backend):
    def _init(self, server, topic):
        self.server = server
        self.topic = topic
        self.producer = None

    def _on_record(self, record):
        if record.topic != self.topic: return

        try:
            msg = json.loads(record.value.decode('utf-8'))
        except Exception as e:
            logging.exception(e)

        logging.debug('Received message: {}'.format(msg))
        self.on_msg(msg)

    def _init_producer(self):
        if not self.producer:
            self.producer = KafkaProducer(bootstrap_servers=self.server)

    def send_msg(self, msg):
        if isinstance(msg, dict):
            msg = json.dumps(msg)
        if isinstance(msg, str):
            msg = msg.encode('utf-8')
        if not isinstance(msg, bytes):
            msg = json.dumps(msg)
            raise RuntimeError('Invalid non-JSON message')

        self._init_producer()
        self.producer.send(self.topic, msg)
        self.producer.flush()

    def run(self):
        self.consumer = KafkaConsumer(self.topic, bootstrap_servers=self.server)
        for msg in self.consumer:
            self._on_record(msg)

# vim:sw=4:ts=4:et:




import logging
import json
import os
import time

from .. import Backend

class LocalBackend(Backend):
    def _init(self, fifo):
        self.fifo = fifo
        try: os.mkfifo(self.fifo)
        except FileExistsError as e: pass
        logging.info('Initialized local backend on fifo {}'.format(self.fifo))

    def send_msg(self, msg):
        if isinstance(msg, dict):
            msg = json.dumps(msg)
        if not isinstance(msg, str):
            raise RuntimeError('Invalid non-JSON message')

        msglen = len(msg)+1  # Include \n
        msg = bytearray((str(msglen) + '\n' + msg + '\n').encode('utf-8'))
        with open(self.fifo, 'wb') as f:
            f.write(msg)

    def run(self):
        with open(self.fifo, 'rb', 0) as f:
            while True:
                try:
                    msglen = int(f.readline())
                except ValueError as e:
                    time.sleep(0.1)
                    continue

                msg = f.read(msglen-1)
                if not msg: continue

                try:
                    msg = json.loads(msg.decode('utf-8'))
                except Exception as e:
                    logging.exception(e)
                    continue

                logging.debug('Received message: {}'.format(msg))
                self.on_msg(msg)

# vim:sw=4:ts=4:et:




import logging
import time

from phue import Bridge

from .. import LightPlugin

class LightHuePlugin(LightPlugin):
    """ Python dependencies """
    _requires = [
        'phue'
    ]

    MAX_BRI = 255
    MAX_SAT = 255
    MAX_HUE = 65535

    def _init(self):
        self.bridge_address = self.config['bridge']
        self.bridge = None
        logging.info('Initializing Hue lights plugin - bridge: "{}"'.
                     format(self.bridge_address))

        self.connect()
        self.lights = []; self.groups = []

        if 'lights' in self.config:
            self.lights = self.config['lights']
        elif 'groups' in self.config:
            self.groups = self.config['groups']
            self._expand_groups(self.groups)
        else:
            self.lights = [l.name for l in self.bridge.lights]

        logging.info('Configured lights: "{}"'. format(self.lights))

    def _expand_groups(self, group_names):
        groups = [g for g in self.bridge.groups
                if g.name in group_names]

        for g in groups:
            self.lights.extend([l.name for l in g.lights])

    def connect(self):
        # Lazy init
        if not self.bridge:
            self.bridge = Bridge(self.bridge_address)
            logging.info('Bridge connected')

            self.get_scenes()

            # uncomment these lines if you're running huectrl for
            # the first time and you need to pair it to the switch

            # self.bridge.connect()
            # self.bridge.get_api()
        else:
            logging.info('Bridge already connected')


    def get_scenes(self):
        scenes = [s.name for s in self.bridge.scenes]
        # TODO Expand it with custom scenes specified in config.yaml as in #14

    def _execute(self, attr, *args, **kwargs):
        try:
            self.connect()
        except Exception as e:
            logging.exception(e)
            # Reset bridge connection
            self.bridge = None

            if 'is_retry' not in kwargs:
                time.sleep(1)
                self._execute(attr, is_retry=True, *args, **kwargs)

            return

        lights = []; groups = []
        if 'lights' in kwargs and kwargs['lights']:
            lights = kwargs['lights'].split(',') \
                if isinstance(lights, str) else kwargs['lights']
        elif 'groups' in kwargs and kwargs['lights']:
            groups = kwargs['groups'].split(',') \
                if isinstance(groups, str) else kwargs['groups']
        else:
            lights = self.lights
            groups = self.groups

        logging.info('[Attribute: {}] [Values: {}] [Lights: {}] [Groups: {}]'.
                     format(attr, args, lights, groups))

        try:
            if attr == 'scene':
                self.bridge.run_scene(groups[0], kwargs['name'])
            elif groups:
                self.bridge.set_group(groups, attr, *args)
            elif lights:
                self.bridge.set_light(lights, attr, *args)
        except Exception as e:
            logging.exception(e)
            # Reset bridge connection
            self.bridge = None

    def on(self, lights=[], groups=[]):
        self._execute('on', True, lights=lights, groups=groups)

    def off(self, lights=[], groups=[]):
        self._execute('on', False, lights=lights, groups=groups)

    def bri(self, value, lights=[], groups=[]):
        self._execute('bri', int(value) % (self.MAX_BRI+1),
                      lights=lights, groups=groups)

    def sat(self, value, lights=[], groups=[]):
        self._execute('sat', int(value) % (self.MAX_SAT+1),
                      lights=lights, groups=groups)

    def hue(self, value, lights=[], groups=[]):
        self._execute('hue', int(value) % (self.MAX_HUE+1),
                      lights=lights, groups=groups)

    def hue(self, value, lights=[], groups=[]):
        self._execute('hue', int(value) % (self.MAX_HUE+1),
                      lights=lights, groups=groups)

    def scene(self, name, lights=[], groups=[]):
        self._execute('scene', name=name, lights=lights, groups=groups)

    def status(self):
        return ['']

# vim:sw=4:ts=4:et:




import functools
import importlib
import os
import logging
import json
import socket
import sys
import websocket
import yaml

from queue import Queue
from threading import Thread
from getopt import getopt

__author__ = 'Fabio Manganiello <info@fabiomanganiello.com>'

#-----------#

config = {}
modules = {}
wrkdir = os.path.dirname(os.path.realpath(__file__))


def _init_plugin(plugin_name, reload=False):
    global modules
    global config

    if plugin_name in modules and not reload:
        return modules[plugin_name]

    try:
        module = importlib.import_module(__package__ + '.plugins.' + plugin_name)
    except ModuleNotFoundError as e:
        logging.warn('No such plugin: {}'.format(plugin_name))
        raise RuntimeError(e)

    # e.g. plugins.music.mpd main class: MusicMpdPlugin
    cls_name = functools.reduce(
        lambda a,b: a.title() + b.title(),
        (plugin_name.title().split('.'))
    ) + 'Plugin'

    plugin_conf = config[plugin_name] if plugin_name in config else {}

    try:
        plugin = getattr(module, cls_name)(plugin_conf)
        modules[plugin_name] = plugin
    except AttributeError as e:
        logging.warn('No such class in {}: {}'.format(
            plugin_name, cls_name))
        raise RuntimeError(e)

    return plugin


def _exec_func(args, retry=True):
    action = args.pop('action')
    tokens = action.split('.')
    module_name = str.join('.', tokens[:-1])
    method_name = tokens[-1:][0]

    try:
        plugin = _init_plugin(module_name)
    except RuntimeError as e:  # Module/class not found
        logging.exception(e)
        return

    try:
        ret = plugin.run(method=method_name, **args)
        out = None
        err = None

        if isinstance(ret, list):
            out = ret[0]
            err = ret[1] if len(ret) > 1 else None
        elif ret is not None:
            out = ret

        if out:
            logging.info('Command output: {}'.format(out))

        if err:
            logging.warn('Command error: {}'.format(err))
    except Exception as e:
        logging.exception(e)
        if retry:
            # Put the action back where it was before retrying
            args['action'] = action

            logging.info('Reloading plugin {} and retrying'.format(module_name))
            _init_plugin(module_name, reload=True)
            _exec_func(args, retry=False)


def on_msg(msg):
    Thread(target=_exec_func, args=(msg,)).start()


def parse_config_file(config_file=None):
    global config

    if config_file:
        locations = [config_file]
    else:
        locations = [
            # ./config.yaml
            os.path.join(wrkdir, 'config.yaml'),
            # ~/.config/platypush/config.yaml
            os.path.join(os.environ['HOME'], '.config', 'platypush', 'config.yaml'),
            # /etc/platypush/config.yaml
            os.path.join(os.sep, 'etc', 'platypush', 'config.yaml'),
        ]

    for loc in locations:
        try:
            with open(loc,'r') as f:
                config = yaml.load(f)
        except FileNotFoundError as e:
            pass

    for section in config:
        if 'disabled' in config[section] and config[section]['disabled']:
            del config[section]

    return config


def get_backends(config):
    backends = []

    for k in config.keys():
        if k.startswith('backend.'):
            module = importlib.import_module(__package__ + '.' + k)

            # e.g. backend.pushbullet main class: PushbulletBackend
            cls_name = functools.reduce(
                lambda a,b: a.title() + b.title(),
                (module.__name__.title().split('.')[2:])
            ) + 'Backend'

            # Ignore the pusher attribute here
            if 'pusher' in config[k]: del config[k]['pusher']

            try:
                b = getattr(module, cls_name)(config[k])
                backends.append(b)
            except AttributeError as e:
                logging.warn('No such class in {}: {}'.format(
                    module.__name__, cls_name))
                raise RuntimeError(e)

    return backends

def get_device_id():
    global config
    return config['device_id']


def main():
    DEBUG = False
    config_file = None

    plugins_dir = os.path.join(wrkdir, 'plugins')
    sys.path.insert(0, plugins_dir)

    optlist, args = getopt(sys.argv[1:], 'vh')
    for opt, arg in optlist:
        if opt == '-c':
            config_file = arg
        if opt == '-v':
            DEBUG = True
        elif opt == '-h':
            print('''
Usage: {} [-v] [-h] [-c <config_file>]
    -v  Enable debug mode
    -h  Show this help
    -c  Path to the configuration file (default: ./config.yaml)
'''.format(sys.argv[0]))
            return

    config = parse_config_file(config_file)
    logging.info('Configuration dump: {}'.format(config))

    if 'device_id' not in config:
        config['device_id'] = socket.gethostname()

    if 'debug' in config:
        DEBUG = config['debug']

    if DEBUG:
        logging.basicConfig(level=logging.DEBUG)
        websocket.enableTrace(True)
    else:
        logging.basicConfig(level=logging.INFO)

    mq = Queue()
    backends = get_backends(config)

    for backend in backends:
        backend.mq = mq
        backend.start()

    while True:
        try:
            on_msg(mq.get())
        except KeyboardInterrupt:
            return

if __name__ == '__main__':
    main()

# vim:sw=4:ts=4:et:




import logging

from ouimeaux.environment import Environment, UnknownDevice

from .. import SwitchPlugin

class SwitchWemoPlugin(SwitchPlugin):
    _requires = [
        'ouimeaux'
    ]

    def _init(self):
        logging.basicConfig(level=logging.INFO)

        self.env = Environment()
        self.env.start()
        logging.info('Starting WeMo discovery')
        self.env.discover(seconds=3)

    def on(self, device):
        switch = self.env.get_switch(device)
        logging.info('Turning {} on'.format(device))
        switch.on()

    def off(self, device):
        switch = self.env.get_switch(device)
        logging.info('Turning {} off'.format(device))
        switch.off()

    def toggle(self, device):
        switch = self.env.get_switch(device)
        logging.info('Toggling {}'.format(device))
        switch.toggle()

    def status(self):
        return ['']

# vim:sw=4:ts=4:et:




import logging
import platypush

from threading import Thread

def _default_on_init(backend):
    logging.info('Backend {} initialized'.format(backend.__module__))


def _default_on_close(backend):
    logging.info('Backend {} terminated'.format(backend.__module__))


def _default_on_msg(backend, msg):
    logging.info('Received message: {}'.format(msg))


def _default_on_error(backend, error):
    logging.error(error)


class Backend(Thread):
    def __init__(self, config, mq = None,
                 on_init  = _default_on_init,
                 on_close = _default_on_close,
                 on_error = _default_on_error):
        self.config = config
        self.mq = mq
        self.on_init = on_init
        self.on_close = on_close
        self.on_error = on_error

        Thread.__init__(self)
        logging.basicConfig(level=logging.INFO
                            if 'logging' not in config
                            else getattr(logging, config.pop('logging')))

        for cls in reversed(self.__class__.mro()):
            if cls is not object and hasattr(cls, '_init'):
                cls._init(self, **config)

    def is_local(self):
        from platypush.backend.local import LocalBackend
        return isinstance(self, LocalBackend)

    def on_msg(self, msg):
        if 'target' not in msg:
            return  # No target

        target = msg.pop('target')
        if target != platypush.get_device_id() and not self.is_local():
            return  # Not for me

        if 'action' not in msg:
            self.on_error('No action specified: {}'.format(msg))
            return

        self.mq.put(msg)

    def send_msg(self, msg):
        raise NotImplementedError()

    def run(self):
        raise NotImplementedError()

# vim:sw=4:ts=4:et:




import os
import sys
import logging

class Plugin(object):
    def __init__(self, config):
        self.config = config
        self._set_logging()

        for cls in reversed(self.__class__.mro()):
            if cls is not object:
                try:
                    cls._init(self)
                except AttributeError as e:
                    pass

    def _set_logging(self):
        if 'logging' in self.config:
            logging.basicConfig(level=getattr(logging, self.config['logging']))
        else:
            logging.basicConfig(level=logging.INFO)

    def run(self, method, *args, **kwargs):
        res = getattr(self, method)(*args, **kwargs)
        return res

# vim:sw=4:ts=4:et:




import mpd

from .. import MusicPlugin

class MusicMpdPlugin(MusicPlugin):
    _requires = [
        'mpd'
    ]

    def _init(self):
        self.client = mpd.MPDClient(use_unicode=True)
        self.client.connect(self.config['host'], self.config['port'])

    def play(self):
        self.client.play()

    def pause(self):
        self.client.pause()

    def stop(self):
        self.client.stop()

    def next(self):
        self.client.next()

    def previous(self):
        self.client.previous()

    def setvol(self, vol):
        self.client.setvol(vol)

    def add(self, content):
        self.client.add(content)

    def playlistadd(self, playlist):
        self.client.playlistadd(playlist)

    def clear(self):
        self.client.clear()

    def status(self):
        return self.client.status()

# vim:sw=4:ts=4:et:




import logging
import json
import websocket

from pushbullet import Pushbullet

from .. import Backend

class PushbulletBackend(Backend):
    _requires = [
        'pushbullet'
    ]

    def _init(self, token, device):
        self.token = token
        self.device = device

    @staticmethod
    def _on_init(ws):
        logging.info('Connection opened')

    @staticmethod
    def _on_close(ws):
        logging.info('Connection closed')

    @staticmethod
    def _on_msg(ws, msg):
        ws.backend._on_push(msg)

    @staticmethod
    def _on_error(ws, e):
        logging.exception(e)
        self.ws.close()
        self._init_socket()

    def _on_push(self, data):
        try:
            data = json.loads(data) if isinstance(data, str) else push
        except Exception as e:
            logging.exception(e)
            return

        if data['type'] != 'push':
            return  # Not a push notification

        push = data['push']
        logging.debug('Received push: {}'.format(push))

        if 'body' not in push:
            return

        body = push['body']
        try:
            body = json.loads(body)
        except ValueError as e:
            return

        self.on_msg(body)

    def _init_socket(self):
        self.ws = websocket.WebSocketApp(
            'wss://stream.pushbullet.com/websocket/' + self.token,
            on_open = self._on_init,
            on_message = self._on_msg,
            on_error = self._on_error,
            on_close = self._on_close)

        self.ws.backend = self

    def send_msg(self, msg):
        if isinstance(msg, dict):
            msg = json.dumps(msg)
        if not isinstance(msg, str):
            raise RuntimeError('Invalid non-JSON message')

        pb = Pushbullet(self.token)
        devices = [dev for dev in pb.devices if dev.nickname == self.device]
        if not devices:
            raise RuntimeError('No such device: {}'.format(self.device))

        device = devices[0]
        pb.push_note('', msg, device)

    def run(self):
        self._init_socket()
        self.ws.run_forever()

# vim:sw=4:ts=4:et:

